import { __decorate, __metadata } from 'tslib';
import { Actions, Effect } from '@ngrx/effects';
import { distinctUntilChanged, filter, map, debounceTime, skip, switchMap, take } from 'rxjs/operators';
import { merge, Subject, of, Observable } from 'rxjs';
import { CommonModule } from '@angular/common';
import { FormGroup, FormGroupDirective, FormsModule, ReactiveFormsModule } from '@angular/forms';
import { Store } from '@ngrx/store';
import { InjectionToken, Inject, Injectable, Optional, Directive, Input, Self, ElementRef, SkipSelf, NgModule } from '@angular/core';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @return {?}
 */
function noStoreError() {
    throw new Error(`No store detected. Make sure you've included @ngrx/store in your application.`);
}
/**
 * @return {?}
 */
function noFormGroupError() {
    throw new Error(`rxfConnectForm needs to be placed on an element which also contains a formGroup directive.`);
}
/**
 * @param {?} path
 * @return {?}
 */
function noStoreFormBinding(path) {
    throw new Error(`No binding found for form state on path '${path}'.`);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const STORE_FORMS_CONFIG = new InjectionToken('rxsf Config');
/** @type {?} */
const STORE_FORMS_FEATURE = new InjectionToken('rxsf Feature');

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const defaultStoreFormsConfig = {
    bindingStrategy: 'ObserveStore',
    debounce: 100,
    errorMessages: {}
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} object
 * @param {?} path
 * @param {?=} throwOnMiss
 * @return {?}
 */
function deepGet(object, path, throwOnMiss = false) {
    return path
        .split('.')
        .reduce((partialState, pathSegment) => {
        if (throwOnMiss && typeof partialState[pathSegment] !== 'object') {
            throw new Error(`Path '${path}' could not be resolved in object ${object}. Segment '${pathSegment}' was not referring to type object.`);
        }
        return partialState[pathSegment] || {};
    }, object);
}
/**
 * @param {?} userConfig
 * @return {?}
 */
function getEffectiveConfig(userConfig) {
    return Object.assign({}, defaultStoreFormsConfig, userConfig);
}
/**
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function deepEquals(x, y) {
    if (x === y) {
        return true; // if both x and y are null or undefined and exactly the same
    }
    else if (!(x instanceof Object) || !(y instanceof Object)) {
        return false; // if they are not strictly equal, they both need to be Objects
    }
    else if (x.constructor !== y.constructor) {
        // they must have the exact same prototype chain, the closest we can do is
        // test their constructor.
        return false;
    }
    else {
        for (const p in x) {
            if (!x.hasOwnProperty(p)) {
                continue; // other properties were tested using x.constructor === y.constructor
            }
            if (!y.hasOwnProperty(p)) {
                return false; // allows to compare x[ p ] and y[ p ] when set to undefined
            }
            if (x[p] === y[p]) {
                continue; // if they have the same strict value or identity then they are equal
            }
            if (typeof (x[p]) !== 'object') {
                return false; // Numbers, Strings, Functions, Booleans must be strictly equal
            }
            if (!deepEquals(x[p], y[p])) {
                return false;
            }
        }
        for (const p in y) {
            if (y.hasOwnProperty(p) && !x.hasOwnProperty(p)) {
                return false;
            }
        }
        return true;
    }
}
/**
 * @param {?} control
 * @return {?}
 */
function extractStateBase(control) {
    return {
        dirty: control.dirty,
        invalid: control.invalid,
        pending: control.pending,
        pristine: control.pristine,
        touched: control.touched,
        untouched: control.untouched,
        valid: control.valid,
        disabled: control.disabled,
        enabled: control.enabled
    };
}
// TODO: Handle form arrays correctly
/**
 * @template F
 * @param {?} control
 * @param {?} errorResolver
 * @param {?=} path
 * @return {?}
 */
function getFormState(control, errorResolver, path = []) {
    if (control instanceof FormGroup) {
        /** @type {?} */
        const fields = Object.keys(control.controls)
            .map(key => ({
            name: key,
            state: getFormState(control.controls[key], errorResolver, [...path, key])
        }))
            .reduce((reducedFields, entry) => {
            reducedFields[entry.name] = entry.state;
            return reducedFields;
        }, (/** @type {?} */ ({})));
        return (/** @type {?} */ (Object.assign({}, extractStateBase(control), { value: (/** @type {?} */ (control.value)), fields, errors: errorResolver(control, path) })));
    }
    else {
        return (/** @type {?} */ (Object.assign({}, extractStateBase(control), { value: control.value, errors: errorResolver(control, path) })));
    }
}
/**
 * @param {?} errors
 * @param {?} path
 * @param {?=} errorMessages
 * @param {?=} pathPrefix
 * @return {?}
 */
function resolveErrors(errors, path, errorMessages = {}, pathPrefix = []) {
    if (!errors) {
        return null;
    }
    /** @type {?} */
    const normalizedPath = path.filter(pathElement => !(/^[\d\s]+$/g.test(pathElement)));
    return Object.keys(errors)
        .reduce((messages, validatorName) => {
        /** @type {?} */
        const wholePath = [...pathPrefix, ...normalizedPath];
        /** @type {?} */
        let resolvedMessage = deepGet(errorMessages, [...wholePath, validatorName].join('.'));
        if (typeof resolvedMessage === 'object') {
            /** @type {?} */
            const parentSearchPath = [...wholePath];
            while (typeof resolvedMessage === 'object' && parentSearchPath.length > 0) {
                parentSearchPath.pop();
                resolvedMessage = deepGet(errorMessages, [...parentSearchPath, validatorName].join('.'));
            }
            if (typeof resolvedMessage === 'object') {
                resolvedMessage = validatorName;
            }
        }
        if (messages.indexOf(resolvedMessage) === -1) {
            messages.push(resolvedMessage);
        }
        return messages;
    }, []);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template F
 */
class UpdateStoreFormStateAction {
    /**
     * @param {?} path
     * @param {?} formState
     */
    constructor(path, formState) {
        this.path = path;
        this.formState = formState;
        this.type = 'UpdateStoreFormStateAction';
    }
}
/**
 * @template F
 */
class UpdateStoreFormAction {
    /**
     * @param {?} path
     * @param {?} value
     */
    constructor(path, value) {
        this.path = path;
        this.value = value;
        this.type = 'UpdateStoreFormAction';
    }
}
class StoreFormsNoopAction {
    constructor() {
        this.type = 'StoreFormsNoopAction';
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class StoreFormsService {
    /**
     * @param {?} config
     * @param {?} store
     * @param {?} feature
     */
    constructor(config, store, feature) {
        this.store = store;
        this.feature = feature;
        this.bindings = {};
        this.pathPrefix = '';
        this.config = getEffectiveConfig(config);
        if (!this.store) {
            noStoreError();
        }
        if (this.feature) {
            this.pathPrefix = `${feature}.`;
        }
    }
    /**
     * @param {?} path
     * @param {?} formGroup
     * @return {?}
     */
    bind(path, formGroup) {
        /** @type {?} */
        const stateUpdateTrigger = new Subject();
        /** @type {?} */
        const errorResolver = (control, controlPath) => resolveErrors(control.errors, controlPath, this.config.errorMessages, pathWithPrefix.split('.'));
        /** @type {?} */
        const pathWithPrefix = `${this.pathPrefix}${path}`;
        /** @type {?} */
        let observeStore = true;
        /** @type {?} */
        let source = merge(formGroup.valueChanges.pipe(map(() => getFormState(formGroup, errorResolver))), formGroup.statusChanges.pipe(map(() => getFormState(formGroup, errorResolver))), stateUpdateTrigger.pipe(map(() => getFormState(formGroup, errorResolver))));
        if (this.config.debounce) {
            source = source.pipe(debounceTime(this.config.debounce));
        }
        /** @type {?} */
        const formGroupSubscription = source
            .pipe(distinctUntilChanged((a, b) => deepEquals(a, b)))
            .subscribe(() => {
            observeStore = false;
            this.store.dispatch(new UpdateStoreFormStateAction(pathWithPrefix, (/** @type {?} */ (getFormState(formGroup, errorResolver)))));
            observeStore = true;
        });
        this.bindings[pathWithPrefix] = {
            path: pathWithPrefix,
            formGroup,
            formGroupSubscription,
            stateUpdateTrigger
        };
        if (this.config.bindingStrategy === 'ObserveStore') {
            this.bindings[pathWithPrefix].storeSubscription = this.store
                .pipe(
            // We skip the initial value emitted by the store subject
            skip(1), map((state) => deepGet(state, pathWithPrefix)), filter(() => observeStore), filter((formState) => {
                // Very simple dirty checking by comparing current values in
                // state to values in form group using deep equal to prevent
                // unwanted infinite loop
                return formState.value && Object.keys(formState.value).length && !deepEquals(formState, getFormState(formGroup, errorResolver));
            }))
                .subscribe((formState) => {
                formGroup.setValue(formState.value);
            });
        }
    }
    /**
     * @param {?} path
     * @return {?}
     */
    unbind(path) {
        /** @type {?} */
        const pathWithPrefix = `${this.pathPrefix}${path}`;
        if (!this.bindings[pathWithPrefix]) {
            return;
        }
        this.bindings[pathWithPrefix].formGroupSubscription.unsubscribe();
        if (this.bindings[pathWithPrefix].storeSubscription) {
            this.bindings[pathWithPrefix].storeSubscription.unsubscribe();
        }
        delete this.bindings[pathWithPrefix];
    }
    /**
     * @param {?} path
     * @param {?} replace
     * @return {?}
     */
    replaceBinding(path, replace) {
        this.unbind(path);
        this.bind(path, replace);
    }
    /**
     * @param {?} path
     * @return {?}
     */
    findBinding(path) {
        /** @type {?} */
        const pathWithPrefix = `${this.pathPrefix}${path}`;
        return this.bindings[pathWithPrefix] || null;
    }
    /**
     * @param {?} path
     * @return {?}
     */
    getFormGroup(path) {
        /** @type {?} */
        const binding = this.findBinding(path);
        if (binding) {
            return binding;
        }
        else {
            noStoreFormBinding(`${this.pathPrefix}${path}`);
        }
    }
    /**
     * @param {?} path
     * @return {?}
     */
    triggerUpdate(path) {
        /** @type {?} */
        const binding = this.findBinding(path);
        if (binding) {
            binding.stateUpdateTrigger.next();
        }
        else {
            noStoreFormBinding(`${this.pathPrefix}${path}`);
        }
    }
    /**
     * @param {?} path
     * @param {?} value
     * @return {?}
     */
    updateFormGroup(path, value) {
        /** @type {?} */
        const pathWithPrefix = `${this.pathPrefix}${path}`;
        if (!this.bindings[pathWithPrefix]) {
            noStoreFormBinding(pathWithPrefix);
        }
        /** @type {?} */
        const binding = this.bindings[pathWithPrefix];
        binding.formGroup.setValue(value);
    }
}
StoreFormsService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
StoreFormsService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [STORE_FORMS_CONFIG,] }] },
    { type: Store, decorators: [{ type: Optional }] },
    { type: String, decorators: [{ type: Inject, args: [STORE_FORMS_FEATURE,] }, { type: Optional }] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindingDirective {
    /**
     * @param {?} formGroupDirective
     * @param {?} store
     * @param {?} storeFormsService
     */
    constructor(formGroupDirective, store, storeFormsService) {
        this.formGroupDirective = formGroupDirective;
        this.store = store;
        this.storeFormsService = storeFormsService;
        if (!this.formGroupDirective) {
            noFormGroupError();
        }
        if (!this.store) {
            noStoreError();
        }
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        if (changes.formGroup) {
            this.storeFormsService.replaceBinding(this.rxsfBinding, this.formGroup);
        }
        else if (changes.rxsfBinding) {
            this.storeFormsService.unbind(changes.rxsfBinding.previousValue);
            this.storeFormsService.replaceBinding(this.rxsfBinding, this.formGroup);
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.storeFormsService.unbind(this.rxsfBinding);
    }
}
BindingDirective.decorators = [
    { type: Directive, args: [{
                selector: '[rxsfBinding]'
            },] }
];
/** @nocollapse */
BindingDirective.ctorParameters = () => [
    { type: FormGroupDirective, decorators: [{ type: Self }, { type: Optional }] },
    { type: Store, decorators: [{ type: Optional }] },
    { type: StoreFormsService }
];
BindingDirective.propDecorators = {
    rxsfBinding: [{ type: Input }],
    formGroup: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class StoreFormsEffects {
    /**
     * @param {?} actions
     * @param {?} bindingService
     * @param {?} config
     * @param {?} store
     */
    constructor(actions, bindingService, config, store) {
        this.actions = actions;
        this.bindingService = bindingService;
        this.store = store;
        this.reflectToFormStateEffect = this.actions
            .pipe(filter((action) => action instanceof UpdateStoreFormAction), switchMap((action) => {
            if (this.config.bindingStrategy === 'ObserveStore') {
                return (/** @type {?} */ (this.store
                    .pipe(take(1), map((state) => {
                    /** @type {?} */
                    const formState = deepGet(state, action.path, true);
                    return new UpdateStoreFormStateAction(action.path, Object.assign({}, formState, { value: Object.assign({}, formState.value, action.value) }));
                }))));
            }
            else {
                this.bindingService.updateFormGroup(action.path, action.value);
                return of(new StoreFormsNoopAction());
            }
        }));
        this.config = getEffectiveConfig(config);
    }
}
StoreFormsEffects.decorators = [
    { type: Injectable }
];
/** @nocollapse */
StoreFormsEffects.ctorParameters = () => [
    { type: Actions },
    { type: StoreFormsService },
    { type: undefined, decorators: [{ type: Inject, args: [STORE_FORMS_CONFIG,] }] },
    { type: Store }
];
__decorate([
    Effect(),
    __metadata("design:type", Observable)
], StoreFormsEffects.prototype, "reflectToFormStateEffect", void 0);

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class TouchedUpdateDirective {
    /**
     * @param {?} elementRef
     * @param {?} binding
     * @param {?} storeFormsService
     */
    constructor(elementRef, binding, storeFormsService) {
        this.elementRef = elementRef;
        this.binding = binding;
        this.storeFormsService = storeFormsService;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        if (!this.binding) {
            return;
        }
        this.blurListener = () => this.storeFormsService.triggerUpdate(this.binding.rxsfBinding);
        this.elementRef.nativeElement.addEventListener('blur', this.blurListener);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.blurListener) {
            this.elementRef.nativeElement.removeEventListener('blur', this.blurListener);
        }
    }
}
TouchedUpdateDirective.decorators = [
    { type: Directive, args: [{
                // tslint:disable-next-line
                selector: 'input[formControlName],input[formControl]'
            },] }
];
/** @nocollapse */
TouchedUpdateDirective.ctorParameters = () => [
    { type: ElementRef },
    { type: BindingDirective, decorators: [{ type: SkipSelf }, { type: Optional }] },
    { type: StoreFormsService }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
const ɵ0 = defaultStoreFormsConfig;
class StoreFormsModule {
    /**
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config) {
        return {
            ngModule: StoreFormsModule,
            providers: [{
                    provide: STORE_FORMS_CONFIG,
                    useValue: config
                }, {
                    provide: StoreFormsService,
                    useClass: StoreFormsService,
                    deps: [STORE_FORMS_CONFIG, Store]
                }]
        };
    }
    /**
     * @param {?} feature
     * @param {?=} config
     * @return {?}
     */
    static forFeature(feature, config) {
        return {
            ngModule: StoreFormsModule,
            providers: [{
                    provide: STORE_FORMS_CONFIG,
                    useValue: config
                }, {
                    provide: StoreFormsService,
                    useClass: StoreFormsService,
                    deps: [STORE_FORMS_CONFIG, Store, STORE_FORMS_FEATURE]
                }, {
                    provide: STORE_FORMS_FEATURE,
                    useValue: feature
                }]
        };
    }
}
StoreFormsModule.decorators = [
    { type: NgModule, args: [{
                imports: [
                    CommonModule,
                    FormsModule,
                    ReactiveFormsModule
                ],
                declarations: [
                    BindingDirective,
                    TouchedUpdateDirective
                ],
                exports: [
                    BindingDirective,
                    TouchedUpdateDirective
                ],
                providers: [{
                        provide: STORE_FORMS_CONFIG,
                        useValue: ɵ0
                    }, {
                        provide: StoreFormsService,
                        useClass: StoreFormsService,
                        deps: [STORE_FORMS_CONFIG, Store]
                    }]
            },] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} state
 * @param {?} action
 * @return {?}
 */
function storeFormsReducer(state, action) {
    switch (action.type) {
        case 'UpdateStoreFormStateAction': {
            /** @type {?} */
            const clone = Object.assign({}, state);
            /** @type {?} */
            const formState = (/** @type {?} */ (action.path
                .split('.')
                .reduce((partialState, pathSegment) => {
                return partialState[pathSegment] = Object.assign({}, partialState[pathSegment]);
            }, clone)));
            Object.assign(formState, Object.assign({}, action.formState));
            if (action.formState.value) {
                Object.assign(formState, {
                    value: Object.assign({}, formState.value, action.formState.value)
                });
            }
            return clone;
        }
    }
    return state;
}
/**
 * @param {?} reducer
 * @return {?}
 */
function storeFormsMetaReducer(reducer) {
    return function (state, action) {
        /** @type {?} */
        const updatedState = storeFormsReducer(state, action);
        return reducer(updatedState, action);
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { UpdateStoreFormAction, BindingDirective, StoreFormsEffects, StoreFormsModule, storeFormsReducer, storeFormsMetaReducer, StoreFormsService, STORE_FORMS_CONFIG, STORE_FORMS_FEATURE, defaultStoreFormsConfig as ɵb, TouchedUpdateDirective as ɵa };

//# sourceMappingURL=ngrx-store-forms.js.map