import { Actions, Effect } from '@ngrx/effects';
import { distinctUntilChanged, filter, map, debounceTime, skip, switchMap, take } from 'rxjs/operators';
import { merge, Subject, of, Observable } from 'rxjs';
import { CommonModule } from '@angular/common';
import { FormGroup, FormGroupDirective, FormsModule, ReactiveFormsModule } from '@angular/forms';
import { Store } from '@ngrx/store';
import { InjectionToken, Inject, Injectable, Optional, Directive, Input, Self, ElementRef, SkipSelf, NgModule } from '@angular/core';
import { __assign, __decorate, __metadata, __spread } from 'tslib';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @return {?}
 */
function noStoreError() {
    throw new Error("No store detected. Make sure you've included @ngrx/store in your application.");
}
/**
 * @return {?}
 */
function noFormGroupError() {
    throw new Error("rxfConnectForm needs to be placed on an element which also contains a formGroup directive.");
}
/**
 * @param {?} path
 * @return {?}
 */
function noStoreFormBinding(path) {
    throw new Error("No binding found for form state on path '" + path + "'.");
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var STORE_FORMS_CONFIG = new InjectionToken('rxsf Config');
/** @type {?} */
var STORE_FORMS_FEATURE = new InjectionToken('rxsf Feature');

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var defaultStoreFormsConfig = {
    bindingStrategy: 'ObserveStore',
    debounce: 100,
    errorMessages: {}
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} object
 * @param {?} path
 * @param {?=} throwOnMiss
 * @return {?}
 */
function deepGet(object, path, throwOnMiss) {
    if (throwOnMiss === void 0) { throwOnMiss = false; }
    return path
        .split('.')
        .reduce(function (partialState, pathSegment) {
        if (throwOnMiss && typeof partialState[pathSegment] !== 'object') {
            throw new Error("Path '" + path + "' could not be resolved in object " + object + ". Segment '" + pathSegment + "' was not referring to type object.");
        }
        return partialState[pathSegment] || {};
    }, object);
}
/**
 * @param {?} userConfig
 * @return {?}
 */
function getEffectiveConfig(userConfig) {
    return __assign({}, defaultStoreFormsConfig, userConfig);
}
/**
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function deepEquals(x, y) {
    if (x === y) {
        return true; // if both x and y are null or undefined and exactly the same
    }
    else if (!(x instanceof Object) || !(y instanceof Object)) {
        return false; // if they are not strictly equal, they both need to be Objects
    }
    else if (x.constructor !== y.constructor) {
        // they must have the exact same prototype chain, the closest we can do is
        // test their constructor.
        return false;
    }
    else {
        for (var p in x) {
            if (!x.hasOwnProperty(p)) {
                continue; // other properties were tested using x.constructor === y.constructor
            }
            if (!y.hasOwnProperty(p)) {
                return false; // allows to compare x[ p ] and y[ p ] when set to undefined
            }
            if (x[p] === y[p]) {
                continue; // if they have the same strict value or identity then they are equal
            }
            if (typeof (x[p]) !== 'object') {
                return false; // Numbers, Strings, Functions, Booleans must be strictly equal
            }
            if (!deepEquals(x[p], y[p])) {
                return false;
            }
        }
        for (var p in y) {
            if (y.hasOwnProperty(p) && !x.hasOwnProperty(p)) {
                return false;
            }
        }
        return true;
    }
}
/**
 * @param {?} control
 * @return {?}
 */
function extractStateBase(control) {
    return {
        dirty: control.dirty,
        invalid: control.invalid,
        pending: control.pending,
        pristine: control.pristine,
        touched: control.touched,
        untouched: control.untouched,
        valid: control.valid,
        disabled: control.disabled,
        enabled: control.enabled
    };
}
// TODO: Handle form arrays correctly
/**
 * @template F
 * @param {?} control
 * @param {?} errorResolver
 * @param {?=} path
 * @return {?}
 */
function getFormState(control, errorResolver, path) {
    if (path === void 0) { path = []; }
    if (control instanceof FormGroup) {
        /** @type {?} */
        var fields = Object.keys(control.controls)
            .map(function (key) { return ({
            name: key,
            state: getFormState(control.controls[key], errorResolver, __spread(path, [key]))
        }); })
            .reduce(function (reducedFields, entry) {
            reducedFields[entry.name] = entry.state;
            return reducedFields;
        }, (/** @type {?} */ ({})));
        return (/** @type {?} */ (__assign({}, extractStateBase(control), { value: (/** @type {?} */ (control.value)), fields: fields, errors: errorResolver(control, path) })));
    }
    else {
        return (/** @type {?} */ (__assign({}, extractStateBase(control), { value: control.value, errors: errorResolver(control, path) })));
    }
}
/**
 * @param {?} errors
 * @param {?} path
 * @param {?=} errorMessages
 * @param {?=} pathPrefix
 * @return {?}
 */
function resolveErrors(errors, path, errorMessages, pathPrefix) {
    if (errorMessages === void 0) { errorMessages = {}; }
    if (pathPrefix === void 0) { pathPrefix = []; }
    if (!errors) {
        return null;
    }
    /** @type {?} */
    var normalizedPath = path.filter(function (pathElement) { return !(/^[\d\s]+$/g.test(pathElement)); });
    return Object.keys(errors)
        .reduce(function (messages, validatorName) {
        /** @type {?} */
        var wholePath = __spread(pathPrefix, normalizedPath);
        /** @type {?} */
        var resolvedMessage = deepGet(errorMessages, __spread(wholePath, [validatorName]).join('.'));
        if (typeof resolvedMessage === 'object') {
            /** @type {?} */
            var parentSearchPath = __spread(wholePath);
            while (typeof resolvedMessage === 'object' && parentSearchPath.length > 0) {
                parentSearchPath.pop();
                resolvedMessage = deepGet(errorMessages, __spread(parentSearchPath, [validatorName]).join('.'));
            }
            if (typeof resolvedMessage === 'object') {
                resolvedMessage = validatorName;
            }
        }
        if (messages.indexOf(resolvedMessage) === -1) {
            messages.push(resolvedMessage);
        }
        return messages;
    }, []);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template F
 */
var /**
 * @template F
 */
UpdateStoreFormStateAction = /** @class */ (function () {
    function UpdateStoreFormStateAction(path, formState) {
        this.path = path;
        this.formState = formState;
        this.type = 'UpdateStoreFormStateAction';
    }
    return UpdateStoreFormStateAction;
}());
/**
 * @template F
 */
var  /**
 * @template F
 */
UpdateStoreFormAction = /** @class */ (function () {
    function UpdateStoreFormAction(path, value) {
        this.path = path;
        this.value = value;
        this.type = 'UpdateStoreFormAction';
    }
    return UpdateStoreFormAction;
}());
var StoreFormsNoopAction = /** @class */ (function () {
    function StoreFormsNoopAction() {
        this.type = 'StoreFormsNoopAction';
    }
    return StoreFormsNoopAction;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var StoreFormsService = /** @class */ (function () {
    function StoreFormsService(config, store, feature) {
        this.store = store;
        this.feature = feature;
        this.bindings = {};
        this.pathPrefix = '';
        this.config = getEffectiveConfig(config);
        if (!this.store) {
            noStoreError();
        }
        if (this.feature) {
            this.pathPrefix = feature + ".";
        }
    }
    /**
     * @param {?} path
     * @param {?} formGroup
     * @return {?}
     */
    StoreFormsService.prototype.bind = /**
     * @param {?} path
     * @param {?} formGroup
     * @return {?}
     */
    function (path, formGroup) {
        var _this = this;
        /** @type {?} */
        var stateUpdateTrigger = new Subject();
        /** @type {?} */
        var errorResolver = function (control, controlPath) {
            return resolveErrors(control.errors, controlPath, _this.config.errorMessages, pathWithPrefix.split('.'));
        };
        /** @type {?} */
        var pathWithPrefix = "" + this.pathPrefix + path;
        /** @type {?} */
        var observeStore = true;
        /** @type {?} */
        var source = merge(formGroup.valueChanges.pipe(map(function () { return getFormState(formGroup, errorResolver); })), formGroup.statusChanges.pipe(map(function () { return getFormState(formGroup, errorResolver); })), stateUpdateTrigger.pipe(map(function () { return getFormState(formGroup, errorResolver); })));
        if (this.config.debounce) {
            source = source.pipe(debounceTime(this.config.debounce));
        }
        /** @type {?} */
        var formGroupSubscription = source
            .pipe(distinctUntilChanged(function (a, b) { return deepEquals(a, b); }))
            .subscribe(function () {
            observeStore = false;
            _this.store.dispatch(new UpdateStoreFormStateAction(pathWithPrefix, (/** @type {?} */ (getFormState(formGroup, errorResolver)))));
            observeStore = true;
        });
        this.bindings[pathWithPrefix] = {
            path: pathWithPrefix,
            formGroup: formGroup,
            formGroupSubscription: formGroupSubscription,
            stateUpdateTrigger: stateUpdateTrigger
        };
        if (this.config.bindingStrategy === 'ObserveStore') {
            this.bindings[pathWithPrefix].storeSubscription = this.store
                .pipe(
            // We skip the initial value emitted by the store subject
            skip(1), map(function (state) { return deepGet(state, pathWithPrefix); }), filter(function () { return observeStore; }), filter(function (formState) {
                // Very simple dirty checking by comparing current values in
                // state to values in form group using deep equal to prevent
                // unwanted infinite loop
                return formState.value && Object.keys(formState.value).length && !deepEquals(formState, getFormState(formGroup, errorResolver));
            }))
                .subscribe(function (formState) {
                formGroup.setValue(formState.value);
            });
        }
    };
    /**
     * @param {?} path
     * @return {?}
     */
    StoreFormsService.prototype.unbind = /**
     * @param {?} path
     * @return {?}
     */
    function (path) {
        /** @type {?} */
        var pathWithPrefix = "" + this.pathPrefix + path;
        if (!this.bindings[pathWithPrefix]) {
            return;
        }
        this.bindings[pathWithPrefix].formGroupSubscription.unsubscribe();
        if (this.bindings[pathWithPrefix].storeSubscription) {
            this.bindings[pathWithPrefix].storeSubscription.unsubscribe();
        }
        delete this.bindings[pathWithPrefix];
    };
    /**
     * @param {?} path
     * @param {?} replace
     * @return {?}
     */
    StoreFormsService.prototype.replaceBinding = /**
     * @param {?} path
     * @param {?} replace
     * @return {?}
     */
    function (path, replace) {
        this.unbind(path);
        this.bind(path, replace);
    };
    /**
     * @param {?} path
     * @return {?}
     */
    StoreFormsService.prototype.findBinding = /**
     * @param {?} path
     * @return {?}
     */
    function (path) {
        /** @type {?} */
        var pathWithPrefix = "" + this.pathPrefix + path;
        return this.bindings[pathWithPrefix] || null;
    };
    /**
     * @param {?} path
     * @return {?}
     */
    StoreFormsService.prototype.getFormGroup = /**
     * @param {?} path
     * @return {?}
     */
    function (path) {
        /** @type {?} */
        var binding = this.findBinding(path);
        if (binding) {
            return binding;
        }
        else {
            noStoreFormBinding("" + this.pathPrefix + path);
        }
    };
    /**
     * @param {?} path
     * @return {?}
     */
    StoreFormsService.prototype.triggerUpdate = /**
     * @param {?} path
     * @return {?}
     */
    function (path) {
        /** @type {?} */
        var binding = this.findBinding(path);
        if (binding) {
            binding.stateUpdateTrigger.next();
        }
        else {
            noStoreFormBinding("" + this.pathPrefix + path);
        }
    };
    /**
     * @param {?} path
     * @param {?} value
     * @return {?}
     */
    StoreFormsService.prototype.updateFormGroup = /**
     * @param {?} path
     * @param {?} value
     * @return {?}
     */
    function (path, value) {
        /** @type {?} */
        var pathWithPrefix = "" + this.pathPrefix + path;
        if (!this.bindings[pathWithPrefix]) {
            noStoreFormBinding(pathWithPrefix);
        }
        /** @type {?} */
        var binding = this.bindings[pathWithPrefix];
        binding.formGroup.setValue(value);
    };
    StoreFormsService.decorators = [
        { type: Injectable }
    ];
    /** @nocollapse */
    StoreFormsService.ctorParameters = function () { return [
        { type: undefined, decorators: [{ type: Inject, args: [STORE_FORMS_CONFIG,] }] },
        { type: Store, decorators: [{ type: Optional }] },
        { type: String, decorators: [{ type: Inject, args: [STORE_FORMS_FEATURE,] }, { type: Optional }] }
    ]; };
    return StoreFormsService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var BindingDirective = /** @class */ (function () {
    function BindingDirective(formGroupDirective, store, storeFormsService) {
        this.formGroupDirective = formGroupDirective;
        this.store = store;
        this.storeFormsService = storeFormsService;
        if (!this.formGroupDirective) {
            noFormGroupError();
        }
        if (!this.store) {
            noStoreError();
        }
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    BindingDirective.prototype.ngOnChanges = /**
     * @param {?} changes
     * @return {?}
     */
    function (changes) {
        if (changes.formGroup) {
            this.storeFormsService.replaceBinding(this.rxsfBinding, this.formGroup);
        }
        else if (changes.rxsfBinding) {
            this.storeFormsService.unbind(changes.rxsfBinding.previousValue);
            this.storeFormsService.replaceBinding(this.rxsfBinding, this.formGroup);
        }
    };
    /**
     * @return {?}
     */
    BindingDirective.prototype.ngOnDestroy = /**
     * @return {?}
     */
    function () {
        this.storeFormsService.unbind(this.rxsfBinding);
    };
    BindingDirective.decorators = [
        { type: Directive, args: [{
                    selector: '[rxsfBinding]'
                },] }
    ];
    /** @nocollapse */
    BindingDirective.ctorParameters = function () { return [
        { type: FormGroupDirective, decorators: [{ type: Self }, { type: Optional }] },
        { type: Store, decorators: [{ type: Optional }] },
        { type: StoreFormsService }
    ]; };
    BindingDirective.propDecorators = {
        rxsfBinding: [{ type: Input }],
        formGroup: [{ type: Input }]
    };
    return BindingDirective;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var StoreFormsEffects = /** @class */ (function () {
    function StoreFormsEffects(actions, bindingService, config, store) {
        var _this = this;
        this.actions = actions;
        this.bindingService = bindingService;
        this.store = store;
        this.reflectToFormStateEffect = this.actions
            .pipe(filter(function (action) { return action instanceof UpdateStoreFormAction; }), switchMap(function (action) {
            if (_this.config.bindingStrategy === 'ObserveStore') {
                return (/** @type {?} */ (_this.store
                    .pipe(take(1), map(function (state) {
                    /** @type {?} */
                    var formState = deepGet(state, action.path, true);
                    return new UpdateStoreFormStateAction(action.path, __assign({}, formState, { value: __assign({}, formState.value, action.value) }));
                }))));
            }
            else {
                _this.bindingService.updateFormGroup(action.path, action.value);
                return of(new StoreFormsNoopAction());
            }
        }));
        this.config = getEffectiveConfig(config);
    }
    StoreFormsEffects.decorators = [
        { type: Injectable }
    ];
    /** @nocollapse */
    StoreFormsEffects.ctorParameters = function () { return [
        { type: Actions },
        { type: StoreFormsService },
        { type: undefined, decorators: [{ type: Inject, args: [STORE_FORMS_CONFIG,] }] },
        { type: Store }
    ]; };
    __decorate([
        Effect(),
        __metadata("design:type", Observable)
    ], StoreFormsEffects.prototype, "reflectToFormStateEffect", void 0);
    return StoreFormsEffects;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var TouchedUpdateDirective = /** @class */ (function () {
    function TouchedUpdateDirective(elementRef, binding, storeFormsService) {
        this.elementRef = elementRef;
        this.binding = binding;
        this.storeFormsService = storeFormsService;
    }
    /**
     * @return {?}
     */
    TouchedUpdateDirective.prototype.ngOnInit = /**
     * @return {?}
     */
    function () {
        var _this = this;
        if (!this.binding) {
            return;
        }
        this.blurListener = function () { return _this.storeFormsService.triggerUpdate(_this.binding.rxsfBinding); };
        this.elementRef.nativeElement.addEventListener('blur', this.blurListener);
    };
    /**
     * @return {?}
     */
    TouchedUpdateDirective.prototype.ngOnDestroy = /**
     * @return {?}
     */
    function () {
        if (this.blurListener) {
            this.elementRef.nativeElement.removeEventListener('blur', this.blurListener);
        }
    };
    TouchedUpdateDirective.decorators = [
        { type: Directive, args: [{
                    // tslint:disable-next-line
                    selector: 'input[formControlName],input[formControl]'
                },] }
    ];
    /** @nocollapse */
    TouchedUpdateDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: BindingDirective, decorators: [{ type: SkipSelf }, { type: Optional }] },
        { type: StoreFormsService }
    ]; };
    return TouchedUpdateDirective;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var ɵ0 = defaultStoreFormsConfig;
var StoreFormsModule = /** @class */ (function () {
    function StoreFormsModule() {
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    StoreFormsModule.forRoot = /**
     * @param {?=} config
     * @return {?}
     */
    function (config) {
        return {
            ngModule: StoreFormsModule,
            providers: [{
                    provide: STORE_FORMS_CONFIG,
                    useValue: config
                }, {
                    provide: StoreFormsService,
                    useClass: StoreFormsService,
                    deps: [STORE_FORMS_CONFIG, Store]
                }]
        };
    };
    /**
     * @param {?} feature
     * @param {?=} config
     * @return {?}
     */
    StoreFormsModule.forFeature = /**
     * @param {?} feature
     * @param {?=} config
     * @return {?}
     */
    function (feature, config) {
        return {
            ngModule: StoreFormsModule,
            providers: [{
                    provide: STORE_FORMS_CONFIG,
                    useValue: config
                }, {
                    provide: StoreFormsService,
                    useClass: StoreFormsService,
                    deps: [STORE_FORMS_CONFIG, Store, STORE_FORMS_FEATURE]
                }, {
                    provide: STORE_FORMS_FEATURE,
                    useValue: feature
                }]
        };
    };
    StoreFormsModule.decorators = [
        { type: NgModule, args: [{
                    imports: [
                        CommonModule,
                        FormsModule,
                        ReactiveFormsModule
                    ],
                    declarations: [
                        BindingDirective,
                        TouchedUpdateDirective
                    ],
                    exports: [
                        BindingDirective,
                        TouchedUpdateDirective
                    ],
                    providers: [{
                            provide: STORE_FORMS_CONFIG,
                            useValue: ɵ0
                        }, {
                            provide: StoreFormsService,
                            useClass: StoreFormsService,
                            deps: [STORE_FORMS_CONFIG, Store]
                        }]
                },] }
    ];
    return StoreFormsModule;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} state
 * @param {?} action
 * @return {?}
 */
function storeFormsReducer(state, action) {
    switch (action.type) {
        case 'UpdateStoreFormStateAction': {
            /** @type {?} */
            var clone = Object.assign({}, state);
            /** @type {?} */
            var formState = (/** @type {?} */ (action.path
                .split('.')
                .reduce(function (partialState, pathSegment) {
                return partialState[pathSegment] = Object.assign({}, partialState[pathSegment]);
            }, clone)));
            Object.assign(formState, __assign({}, action.formState));
            if (action.formState.value) {
                Object.assign(formState, {
                    value: __assign({}, formState.value, action.formState.value)
                });
            }
            return clone;
        }
    }
    return state;
}
/**
 * @param {?} reducer
 * @return {?}
 */
function storeFormsMetaReducer(reducer) {
    return function (state, action) {
        /** @type {?} */
        var updatedState = storeFormsReducer(state, action);
        return reducer(updatedState, action);
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { UpdateStoreFormAction, BindingDirective, StoreFormsEffects, StoreFormsModule, storeFormsReducer, storeFormsMetaReducer, StoreFormsService, STORE_FORMS_CONFIG, STORE_FORMS_FEATURE, defaultStoreFormsConfig as ɵb, TouchedUpdateDirective as ɵa };

//# sourceMappingURL=ngrx-store-forms.js.map